open Async.Std

(* Serializable value to decide on. *)
type value = int
let string_of_val = string_of_int
let val_of_string = int_of_string

(* Protocol specifying the way messages are sent across networks. *)
module type Protocol = sig
  type event = Vote of int * value (* Vote    (round number, value to be sent)*)
             | Collect of int      (* Collect (round number) *)
             | Decide of value     (* Decide  (value decided) *)
             | Connect (* ignore this, only used to establish connection *)

  (* Represent the message sending across the internet: (server id, event) *)
  type msg   = int * event
  (* a channel the current replica uses to hear from another replica *)
  type reader
  (* a channel the current replica uses to send message to another replica *)
  type writer
  (* [send w msg] writes the message [msg] into the writer channel [w] *)
  val send      : writer -> msg -> unit
  (* [receive re] will return a deferred that will be filled when the next
     message from reader [re] was received. *)
  val receive   : reader -> [`Ok of msg |`Eof] Deferred.t
  (* [to_string msg] will serialize the [msg] to string. *)
  val to_string : msg -> string
end

(* Module that configurate the replica. *)
module type Config = sig
  val  server_id : int             (* Server ID *)
  val  total     : int             (* Total number of servers *)
  val  f         : int             (* Tolerance *)
  val  init_val  : value           (* Initial value of this replica *)
  val  logger    : string -> unit  (* Make a log to standard output. *)
end

module type S = sig
  type reader
  type writer

  (* [run plst ret] will bring up a replica, that connects to other replicas via
  the list of reader and writers [lst]. This replica will conduct the
  2/3 consensus protocol. When it decides, it fills in the Ivar [ret]. Run does
  not necessarily return after the replica decides, because it is important
  that it stays in the network of replicas when others ask for its opinion *)
  val run : (reader * writer) list -> value Ivar.t -> unit Deferred.t
end

module Make (P : Protocol) (C : Config)
     : S with type reader  = P.reader
          and type writer  = P.writer
= struct
  type reader = P.reader
  type writer = P.writer
  (* see description in S.run *)
  let run plst ret : unit Deferred.t = failwith "TODO"
end
